#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "../common/objects.h"
#include "../common/debug.h"
#include "../common/vector.h"
#include "intersect.h"
#include "sphere.h"
#include "cylinder.h"
#include "plane.h"
#include "disc.h"

//#include <dlfcn.h>


#include <pthread.h>
extern pthread_mutex_t mutexor;



#define USE_FUNC_POINTERS
extern scene_data* main_scene;
extern int intersects_calced;
extern int skip_object;

typedef double (*intersect_module)(point* startP, vector* projV, object* sphere, point* pntIntersect);
intersect_module intersect_functions[10];


void setup_intersect()
{
	//void* int_handles[10];
	//char* error;

	intersect_functions[3] = &intersect_cylinder;
	intersect_functions[0] = &intersect_sphere;
	intersect_functions[1] = &intersect_plane;
	intersect_functions[2] = &intersect_disc;
/*	
	int_handles[0] = dlopen("./intersect/sphere.so", RTLD_NOW);
	int_handles[1] = dlopen("./intersect/plane.so", RTLD_NOW);

	error = dlerror();
	if (error)
	{
		printd(QUIET, "%s\n", error);
		exit(1);
	}
	else
	{
		printd(ALERT, "intersect handle for: %s\n", "/somewhere/something.so");
	}

	// this horrible typecast is for c++ compliers
	intersect_functions[0] = (double(*)(point*, vector*, object*, point*))
		dlsym(int_handles[0], "intersect_object");
	intersect_functions[1] = (double(*)(point*, vector*, object*, point*))
		dlsym(int_handles[1], "intersect_object");

	error = dlerror();
	if (error)
	{
		printd(QUIET, "error loading lib: %s\n", error);
		exit(1);
	}
	else
	{
		printd(MESSAGE, "intersect function [%i]\n", 1);
	}
*/
}


#ifdef USE_FUNC_POINTERS



/*
 * searches for intersection for a point along a vector
 * calls other intersection functions for specific objects 
 */
double intersect_single_object(point* origin, vector* dir, object *test_obj, int use_shadows)
{
	double dis = 0;

	intersects_calced++;
	
	//FIXME: rotten hack for deferred renderer
	if(test_obj->id == skip_object)
		return 0;
	
	if(use_shadows && test_obj->no_shadow)
		return 0;
	
	dis = 0;
	
	switch (test_obj->obj_type) 
	{/*
		case SPHERE:
			dis = test_obj->intersect(origin, dir, test_obj, NULL);
			//obj->inverse_map(&intsctP, obj, &deb);
			break;
		case PLANE:
			dis = test_obj->intersect(origin, dir, test_obj, NULL);
			break;
		case DISC:
			dis = test_obj->intersect(origin, dir, test_obj, NULL);
			break;
		case BACKGROUND:
			dis = test_obj->intersect(origin, dir, test_obj, NULL);
			break;
		case CYLINDER:
			dis = test_obj->intersect(origin, dir, test_obj, NULL);
			*/
		default:
			dis = test_obj->intersect(origin, dir, test_obj, NULL);
			break;
	}

	return dis;
}

void intersect_all_objects(point* startP, vector* projV, char shadow, hit_data *hit)
{
	double distanceOld =  INFINITY-1;
	double distanceNew = INFINITY;
	object* oHit = NULL;
	int i;
	
	for(i=0; main_scene->models[i] != NULL; i++)
	{
		//FIXME: rotten hack for deferred renderer
		if(i == skip_object)
			continue;
		
		if(shadow && main_scene->models[i]->no_shadow)
			continue;
		
		distanceNew = intersect_single_object(startP, projV, main_scene->models[i], shadow);	
		if(distanceNew > EPSILON && distanceNew < distanceOld)
		{
			distanceOld = distanceNew;
			oHit = main_scene->models[i];
		}
	}

	hit->distance = distanceOld;
	hit->hit_object = oHit;
}

double intersect_dis(point* startP, vector* projV, char shadow)
{
	hit_data hit;
	intersect_all_objects(startP, projV, shadow, &hit);
	return hit.distance;
}


object* intersect_all_objects_getptr(point* origin, vector* dir, point* hit_point, char use_shadows)  
{
	hit_data hit;
	intersect_all_objects(origin, dir, use_shadows, &hit);
	multiply_vector(hit_point, dir, hit.distance);
	add_vectors(hit_point, origin, hit_point);
	
	return hit.hit_object;
}


vector* get_normal(point* p, object* obj, vector* v)
{
	obj->normal(p, obj, v);
    return v;
}




#else


/*
 * searches for intersection for a point along a vector
 * calls other intersection functions for specific objects 
 */
 /*
 
Old code... not used right now.
May move back to it for faster triangle intersects later.

object* intersect(point* startP, vector* projV, point* pHit)  
{
	//these should be infinity
	double distanceOld =  999999998;
	double distanceNew = 999999999;
	object* oHit = NULL;
	int i;
	double dis = 0;
	point intsctP;

	printd(DEBUG, "{ intersect\n");

	for(i=0; main_scene->models[i] != NULL; i++)
	{
		printd(INSANE, "intersect: object #%i\n", i);
		dis = 0;

		switch (main_scene->models[i]->obj_type) 
		{
			case SPHERE:
				dis = intersect_sphere(startP,projV,
						main_scene->models[i],&intsctP);
			case PLANE:
				dis = intersect_plane(startP, projV,
						main_scene->models[i],&intsctP);
			case DISC:
				dis = intersect_disc(startP, projV,
						main_scene->models[i],&intsctP);
		}
		printd(INSANE, "intersect distance: %f\n", dis);

		if(dis>0)
		{
			distanceNew = distance_between(&intsctP,startP);
			if(distanceNew < distanceOld)
			{
				distanceOld = distanceNew;
				pHit->x = intsctP.x;
				pHit->y = intsctP.y;
				pHit->z = intsctP.z;
				oHit = main_scene->models[i];
				main_scene->intrsct.obj_num = i;
			}
		}
	}

	main_scene->intrsct.distance = distanceOld;
	printd(INSANE, "intersect hit: %i\n", oHit);
	printd(DEBUG, "} intersect\n");
	return oHit;
}



vector* get_normal(point* p, object* obj, vector* v)
{
    printd(DEBUG, "{get_normal\n");
    printd(INSANE, " obj num: %i\n", obj->id);
    switch(obj->obj_type)
    {
        case SPHERE:
            normal_sphere(p, obj, v);
            break;
        case PLANE:
            normal_plane(p, obj, v);
            break;
        case DISC:
            normal_disc(p, obj, v); 
            break;
    }
    printd(DEBUG, "}get_normal\n");
    return v;
}
*/
#endif
