//general stuff
#include "../slim.h"

#include "../parse/slm_parser.h"
#include "../parse/obj_slim_converter.h"
#include "../parse/parser.h"
#include "../common/string_extra.h"
#include "./parse_args.h"
#include "./read_scene.h"
//#include "./load_mods.h"
#include "load.h"
#include "./setup_scene.h"
#include "../common/objects.h"
#include "../common/structs.h"
#include "../common/debug.h"
#include <string.h>
#include <stdlib.h>

#include "../common/vector.h"
#define MAX_PATH_LENGTH 1000

//move this to scene object file later
void setup_generic_scene(scene_data *scene)
{
	//light
	if(scene->lights[0] == NULL)
	{
		scene->lights[0] = make_object(LIGHT_INFINITE, 0);
		set_tri(&scene->lights[0]->pos, 1, 0.75, 0.34);
		normalize(&scene->lights[0]->pos);
		make_color(&scene->lights[0]->amb, 255, 255, 255);
		make_color(&scene->lights[0]->spec, 255, 255, 255);
		make_color(&scene->lights[0]->diff, 255, 255, 255);
		printd(MESSAGE, "Light[%i], type[%i-light] - ok\n",	0, scene->lights[0]->obj_type);
	}

	//misc
	scene->height = 401;
	scene->width = 401;
	scene->max_recurs = 2;
	scene->sub_dis = 16;
	scene->antialiasing = 0;
	scene->sub_var = 10;
	scene->sub_level = 1;
}

void setup_generic_camera(scene_data *scene)
{	
	//camera
	set_tri(&scene->camera->lookP, 0, 0, 0);
	set_tri(&scene->camera->look, 0, 0, 0);
	set_tri(&scene->camera->pos, 20, 20, 20);
	set_tri(&scene->camera->rot, 0, 0, 0);
	set_tri(&scene->camera->up, 0, 0, 1);
	printd(MESSAGE, "setting up camera\n");
	scene->camera->units = 5;
}


int init_scene(int argc, char** argv, scene_data* main_scene)
{
	char path[MAX_PATH_LENGTH];
	
	/* load mods in default dir */

	/* parse the arguemnts */
	printd(2, "parsing the args\n");
	get_scene_path(argc, argv, (char*)&path);
	
	load_scene(path, main_scene);

	return 1;
}

void clean_scene(scene_data *scene)
{
	int i;

	for(i=0; scene->models[i]!=NULL;i++)
	{
		free(scene->models[i]);
		scene->models[i] = NULL;
	}

	for(i=0; scene->lights[i]!=NULL;i++)
	{
		free(scene->lights[i]);
		scene->lights[i] = NULL;
	}
	  
	free(scene->frame_buffer);
	free(scene->back_buffer);
	free(scene->misc_buffer);
	free(scene->intersect_buffer);
}

void setup_scene_names(scene_data *main_scene)
{
	char *dot_index;
	
	//get extension
	dot_index = strrchr(main_scene->filename, '.');
	dot_index++;
	strncpy((char*)&main_scene->extension, dot_index, EXTENSION_LENGTH);
	
	strncpy((char*)&main_scene->scene_name, main_scene->filename, strlen(main_scene->filename) - strlen(main_scene->extension) - 1);
}


int load_scene(char *path, scene_data *main_scene)
{
	char error = 0;

	strncpy(main_scene->filename, path, NAME_LENGTH);
	
	clean_scene(main_scene);
	setup_scene_names(main_scene);

	// load the file into memory
	printd(DEBUG, "reading scene\n");
	if( read_file(main_scene->filename, main_scene->file_content, MAX_FILE_SIZE) )
	{
		printd(DEBUG, "parsing scene\n");
		printd(MESSAGE , "file type is: %s\n", main_scene->extension);

		setup_generic_camera(main_scene);
		
		//TODO: fix this crap
		//TERRIBLE HACK
		//allows the background shader to be set while parsing scene file
		setup_background_object();
	
		if(strequal(main_scene->extension, "slm"))
			if( parse_slm_scene(main_scene) == 0)
				error = 1;
		
		if(strequal(main_scene->extension, "obj"))
		{
			setup_generic_scene(main_scene);
			if( obj_reader_for_slim(main_scene, main_scene->filename) == 0)
				error = 1;
			
			//if we didn't find a magic camera, set one up
			if(&main_scene->camera->lookP.x == 0 && &main_scene->camera->look.y == 0 && &main_scene->camera->look.z == 0)
			{
				//look at center
				//add_vectors(&main_scene->camera->lookP, &extreme_dimensions[0], &extreme_dimensions[1]);
				//multiply_vector(&main_scene->camera->lookP, &main_scene->camera->lookP, 0.5);

				//camera at the corner
				//copy_vector(&main_scene->camera->pos, &extreme_dimensions[1]);
				
				//look vector
				//sub_vectors(&scene->camera->look, &scene->camera->lookP, &scene->camera->pos);
			}
		}
		

	}
	else
		error = 1;

	// this does any pre-render calcs
	printd(2, "setting up scene\n");
	if( setup_scene(main_scene) == 0)
		error = 1;

	if(error)
		return 0;

	return 1;
}
