#include <math.h>
#include <time.h>
#include "../common/vector.h"
#include "shaders.h"
#include "phong.h"
#include "../render/tracer.h"
#include "../common/structs.h"
#include "../common/colors.h"
#include "../common/debug.h"
#include "../common/misc.h"
#include "../intersect/intersect.h"

#include <stdio.h>

#include <pthread.h>
extern pthread_mutex_t mutexor;
//pthread_mutex_lock (&mutexor);
//pthread_mutex_unlock (&mutexor);

extern scene_data* main_scene;
extern clock_t current_time;
int skip_object;

// general_shader {{{
void general_shader(intersect_data *i_data, color* final_color)
{
	color new_color;
	intersect_data new_id;
	
	CHECK_STEPS(i_data->step, main_scene->max_recurs);

	clear_color(final_color);
	skip_object = i_data->obj_num;
	//get_normal(&i_data->intersect, i_data->obj, &i_data->normal);
	//normalize(&i_data->reflect);

	// adjust for rays falling inside the object
//	i_data->intersect.x += i_data->normal.x*0.1;
//	i_data->intersect.y += i_data->normal.y*0.1;
//	i_data->intersect.z += i_data->normal.z*0.1;

	// phong
	fill_in_next_hit_data(&new_id, i_data);
	clear_color(&new_color);
	phong(&new_id, &new_color);
	add_colors(final_color, &new_color);

	// reflection 
	fill_in_next_hit_data(&new_id, i_data);
	clear_color(&new_color);
	if(new_id.obj->glossy > 0.0f && new_id.obj->reflect > 0.0f)
		glossy_reflect_color(&new_id, &new_color);
	else if(new_id.obj->reflect > 0.0f)
		reflect_color(&new_id, &new_color);
	//multiply_color(final_color, (1.0 - new_id.obj->reflect)); 
	multiply_color(&new_color, new_id.obj->reflect);
	multiply_color(final_color, 1 - new_id.obj->reflect);
	add_colors(final_color, &new_color);
	clear_color(&new_color);
	i_data->step--;

	// move the intersection for internal object calculations 
	i_data->intersect.x -= i_data->normal.x*0.05;
	i_data->intersect.y -= i_data->normal.y*0.05;
	i_data->intersect.z -= i_data->normal.z*0.05;
	
	// transparency
	fill_in_next_hit_data(&new_id, i_data);
	clear_color(&new_color);
	trans_color(&new_id, &new_color);
	multiply_color(final_color, (1 - new_id.obj->trans));
	multiply_color(&new_color, new_id.obj->trans);
	add_colors(final_color, &new_color);
	clear_color(&new_color);

	// refraction 
	fill_in_next_hit_data(&new_id, i_data);
	clear_color(&new_color);
	refract_color(&new_id, &new_color);
	add_colors(final_color, &new_color);
	clear_color(&new_color);

	skip_object = -1;
	return;
}

// wave_shader {{{
void wave_shader(intersect_data *i_data, color* final_color)
{
	color new_color;
	double dist = 0.0;
	vector bend;
	intersect_data new_id;
	
	CHECK_STEPS(i_data->step, main_scene->max_recurs);

	clear_color(&new_color);

	// create the intersection data structure 
	get_normal(&i_data->intersect, i_data->obj, &i_data->normal);
	reflect_vector(&i_data->reflect, &i_data->normal, &i_data->proj);
	normalize(&i_data->reflect);

	//setup the bumpy surface
	//copy_vector(&new_id.start, &id->intersect);
	//copy_vector(&new_id.proj, &id->proj);
	dist = distance_between(&i_data->intersect, &i_data->obj->pos);
	make_vector(&bend, &i_data->intersect, &i_data->obj->pos); 

	multiply_vector(&bend, &bend, sin(dist+current_time/TIME_SCALE));
	multiply_vector(&bend, &bend, 0.7);
	add_vectors(&bend, &i_data->normal, &bend);

	// adjust for rays falling inside the object 
	i_data->intersect.x += i_data->normal.x*0.01;
	i_data->intersect.y += i_data->normal.y*0.01;
	i_data->intersect.z += i_data->normal.z*0.01;

	// phong 
	phong(i_data, &new_color);
	add_colors(final_color, &new_color);

	// reflection 
	reflect_color(i_data, &new_color);
	//multiply_color(final_color, (1 - i_data->obj->reflect));
	multiply_color(&new_color, i_data->obj->reflect);
	add_colors(final_color, &new_color);

	// move the intersection for internal object calculations 
	i_data->intersect.x -= i_data->normal.x*0.05;
	i_data->intersect.y -= i_data->normal.y*0.05;
	i_data->intersect.z -= i_data->normal.z*0.05;

	// sintransparency 
	add_vectors(&bend, &i_data->proj, &bend);
	trans_color(i_data, &new_color);
	//sin_filter(i_data, &new_color);
	//multiply_color(final_color, (1 - i_data->obj->trans));
	multiply_color(&new_color, i_data->obj->trans);
	add_colors(final_color, &new_color);

	print_color(INSANE, "color:", final_color);
	printd(DEBUG, "}wave_shader\n");
	return;
}


void distance_shader(intersect_data *i_data, color* final_color)
{
	double dis = main_scene->intrsct.distance;

	final_color->r = 10;
	final_color->g = 10;
	final_color->b = 10;

	if(dis < 20)
		final_color->r = (char)(dis/20*256);
	if(dis > 19 && dis < 200)
		final_color->g = (char)(dis/200*256);
	if(dis > 199 && dis < 2000)
		final_color->b = (char)(dis/2000*256);
}

void object_shader(intersect_data *i_data, color* final_color)
{
	color_from_number(final_color, i_data->obj_num);
}

void normal_shader(intersect_data *i_data, color* final_color)
{
	//FIXME:normal also is calced in ray_common, kill one
	get_normal(&i_data->intersect, i_data->obj, &i_data->normal);
	normalize(&i_data->normal);

	final_color->r = (i_data->normal.x*128+128);
	final_color->g = (i_data->normal.y*128+128);
	final_color->b = (i_data->normal.z*128+128);}

void relative_normal_shader(intersect_data *i_data, color* final_color)
{
	double dot = 0;
	//FIXME:normal also is calced in ray_common, kill one
	get_normal(&i_data->intersect, i_data->obj, &i_data->normal);
	normalize(&i_data->normal);
	normalize(&i_data->proj);
	dot = -dot_product(&i_data->normal, &i_data->proj);

	final_color->r = (dot*128+128);
	final_color->g = (dot*128+128);
	final_color->b = (dot*128+128);
}

void checker_shader(intersect_data *i_data, color* final_color)
{
	color new_color;
	vector dist;

	printd(DEBUG, "{wave_shader\n");

	new_color.r = 0;
	new_color.g = 0;
	new_color.b = 0;

	i_data->step++;
	CHECK_STEPS(i_data->step, main_scene->max_recurs);

	// create the intersection data structure 
	get_normal(&i_data->intersect, i_data->obj, &i_data->normal);
	reflect_vector(&i_data->reflect, &i_data->normal, &i_data->proj);
	normalize(&i_data->reflect);
	make_vector(&dist, &i_data->obj->pos, &i_data->intersect);
	
	// adjust for rays falling inside the object 
	i_data->intersect.x += i_data->normal.x*0.01;
	i_data->intersect.y += i_data->normal.y*0.01;
	i_data->intersect.z += i_data->normal.z*0.01;

	if( (int)(floor(dist.x) + floor(dist.z)) % 2 == 0)
		make_color(final_color, 1,1,1);
	else
		make_color(final_color, 255,255,255);

	// phong
	phong(i_data, &new_color);
	add_colors(final_color, &new_color);

	// reflection
	intersect_data new_id;
	fill_in_next_hit_data(&new_id, i_data);
	clear_color(&new_color);
	if(new_id.obj->glossy > 0.0f && new_id.obj->reflect > 0.0f)
		glossy_reflect_color(&new_id, &new_color);
	else if(new_id.obj->reflect > 0.0f)
		reflect_color(&new_id, &new_color);
	//multiply_color(final_color, (1.0 - new_id.obj->reflect)); 
	multiply_color(&new_color, new_id.obj->reflect);
	multiply_color(final_color, 1 - new_id.obj->reflect);
	add_colors(final_color, &new_color);
	clear_color(&new_color);
}

