#include <math.h>
#include <time.h>
#include <stdlib.h>
#include "shaders.h"
#include "../common/vector.h"
#include "../render/tracer.h"
#include "../common/structs.h"
#include "../common/colors.h"
#include "../common/debug.h"
#include "../common/misc.h"
#include "../intersect/intersect.h"
#include "lighting.h"

extern int DEBUG_LEVEL;
extern clock_t current_time;
extern scene_data* main_scene;

void back_black(intersect_data *id, color *bg)
{
	bg->r = 0;
	bg->g = 0;
	bg->b = 0;
}

void back_gradiant(intersect_data *id, color *bg)
{
	bg->r = (id->pos.x + id->pos.y) / 3;  //gradiant
	bg->g = (id->pos.x + id->pos.y) / 3;
	bg->b = (id->pos.x + id->pos.y) / 3;
}


void purple_gradiant(intersect_data *id, color *bg)
{
	bg->r = (id->pos.x + id->pos.y) / 3 * 0.1;  //gradiant
	bg->g = (id->pos.x + id->pos.y) / 3 * 0;
	bg->b = (id->pos.x + id->pos.y) / 3 * 0.3;
}

void back_mountains(intersect_data *id, color *bg)
{
	#define num_mountains 20
	double cosY;
	double cosX;
	vector y_axis;
	vector x_axis;
	int first_mountain;
	int second_mountain;
	double projHeight = 0.20;
	double mountainHeight;
	double percent;
	double mountains[num_mountains] = {0.23, 0.87, 0.65, 0.11, 0.45, 0.185, 0.48, 0.8684,
		0.4183, 0.3285, 0.973, 0.371, 0.4, 0.7113, 0.99285, 0.37, 0.576, 0.48, 0.013, 0.8856};

	set_tri(&y_axis, 0., 1., 0.);
	set_tri(&x_axis, 1., 0., 0.);
	
	normalize(&id->proj);
	
	//nice sunset gradiant
	cosY = dot_product(&id->proj,&y_axis);
	bg->r = (1-cosY)*255;
	bg->g = (1-cosY)*100;
	bg->b = cosY*255;

	//mountains
	cosX = dot_product(&id->proj,&x_axis);
	cosX += 1; //adjust range to 0-2
	cosX = cosX * num_mountains / 2; //adjust range to 0 - num_mountains

	first_mountain = floor(cosX);
	second_mountain = first_mountain + 1;
	if(second_mountain >= num_mountains)
		second_mountain = 0;
	
	percent = cosX - first_mountain;
	mountainHeight = (1-percent)*mountains[first_mountain] + 
			(percent)*mountains[second_mountain];
	mountainHeight = cosY - mountainHeight * projHeight;
	
	if(mountainHeight < 0)
	{
		bg->r = 75*(1+mountainHeight*5);
		bg->g = 50*(1+mountainHeight*5);
		bg->b = 0;
	}
}

void refract_color(intersect_data *id, color *new_color)
{
	// this shader doesn't work right since the transmitted vector hits the same
	// object again from the inside
	intersect_data new_id;

	CHECK_STEPS(id->step, main_scene->max_recurs);	
	fill_in_next_hit_data(&new_id, id);

	copy_vector(&new_id.start, &id->intersect);

	if(new_id.obj->trans > 0 &&
		(id->obj->refract_index <= 3.5 && id->obj->refract_index > 1) )
	{
		refract_vector(&id->normal, &id->proj, id->obj->refract_index, &new_id.proj);
		new_id.obj_num = trace(&new_id.start, &new_id.intersect, &new_id.proj, new_id.step); 
		new_id.obj = main_scene->models[new_id.obj_num];
		new_id.obj->shader(&new_id, new_color);
	}
}

void shader_shadow_cost(intersect_data *id, color *new_color)
{
	int rays_tried;
	int i = 0;
	color shadowed_color;
	
	for(i=0; main_scene->lights[i]!=NULL; i++)
	{
		id->other_obj_num = i;
		get_shadow(id, &shadowed_color);
		rays_tried = shadowed_color.g;
		rays_tried = rays_tried*(255.0f/SHADOW_RAY_SAMPLES);
		color_set_rgb(&shadowed_color, rays_tried, rays_tried, rays_tried);
		add_colors(new_color, &shadowed_color);
	}
	
	return;
}

void reflect_color(intersect_data *i_data, color *new_color)
{
	intersect_data new_id;
	CHECK_STEPS(i_data->step, main_scene->max_recurs);
	fill_in_next_hit_data(&new_id, i_data);

	if(i_data->obj->reflect > 0)
	{
		copy_vector(&new_id.start, &i_data->intersect);
		reflect_vector(&new_id.proj, &i_data->proj, &i_data->normal);
		
		new_id.obj_num = trace(&new_id.start, &new_id.intersect, &new_id.proj, new_id.step);
		new_id.obj = main_scene->models[new_id.obj_num];
		new_id.obj->shader(&new_id, new_color);
	}
}

void glossy_reflect_color(intersect_data *i_data, color *new_color)
{
#define GLOSSY_SAMPLES 200
	int rays_tried;
	int check_every = 35;
	int next_check = check_every;
	double epsilon = 0.001;
	double reflection_amount;
	double last_reflection_amount = 9999999999.0;
	color glossy_color, sample_color;
	vector perfect_reflection_dir;
	double spread =  i_data->obj->glossy;
	double half_spread = spread / 2.0;
	intersect_data new_id;
	
	CHECK_STEPS(i_data->step, main_scene->max_recurs);	

	//get reflection vector
	if(i_data->obj->reflect > 0)
		reflect_vector(&perfect_reflection_dir, &i_data->proj, &i_data->normal);
	else
		return;

	clear_color(&glossy_color);
	for(rays_tried = 1; rays_tried <= GLOSSY_SAMPLES; rays_tried++)
	{
		fill_in_next_hit_data(&new_id, i_data);
		
		//perturb the reflection
		copy_vector(&new_id.start, &i_data->intersect);
		copy_vector(&new_id.proj, &perfect_reflection_dir);
		new_id.proj.x -= ((double)rand()/RAND_MAX) * spread - half_spread;
		new_id.proj.y -= ((double)rand()/RAND_MAX) * spread - half_spread;
		new_id.proj.z -= ((double)rand()/RAND_MAX) * spread - half_spread;
		normalize(&new_id.proj);

		//intersect and shade		
		new_id.obj_num = trace(&new_id.start, &new_id.intersect, &new_id.proj, new_id.step);
		new_id.obj = main_scene->models[new_id.obj_num];
		new_id.obj->shader(&new_id, &sample_color);
		
		add_colors(&glossy_color, &sample_color);
			
		if(rays_tried == next_check)
		{
			reflection_amount = (glossy_color.r + glossy_color.g + glossy_color.b) / (double)rays_tried;
			if(last_reflection_amount == reflection_amount)
				break;

			if(fabs(last_reflection_amount - reflection_amount) <  epsilon)
				break;
			last_reflection_amount = reflection_amount;
			next_check += check_every;
		}
	}

	color_copy(new_color, &glossy_color);
	multiply_color(new_color, 1.0f/((double)(rays_tried-1)));
}

void trans_color(intersect_data *id, color *new_color)
{
	intersect_data new_id;

	CHECK_STEPS(id->step, main_scene->max_recurs);
	fill_in_next_hit_data(&new_id, id);

	if(id->obj->trans > 0)
	{
		copy_vector(&new_id.start, &id->intersect);
		copy_vector(&new_id.proj, &id->proj);
		
		new_id.obj_num = trace(&new_id.start, &new_id.intersect, &new_id.proj, new_id.step);
		new_id.obj = main_scene->models[new_id.obj_num];
		new_id.obj->shader(&new_id, new_color);
	}
	printd(DEBUG, "}trans_color\n");
}

void black_white_filter(intersect_data *id, color *new_color)
{
	intersect_data new_id;
	double bw_color = 0.0;

	CHECK_STEPS(id->step, main_scene->max_recurs);
	fill_in_next_hit_data(&new_id, id);
	
	copy_vector(&new_id.start, &id->intersect);
	copy_vector(&new_id.proj, &id->proj);
	new_id.start.x -= id->normal.x*0.01;
	new_id.start.y -= id->normal.y*0.01;
	new_id.start.z -= id->normal.z*0.01;

	new_id.obj_num = trace(&new_id.start, &new_id.intersect, &new_id.proj, new_id.step);
	new_id.obj = main_scene->models[new_id.obj_num];
	new_id.obj->shader(&new_id, new_color);

	bw_color = new_color->r + new_color->g + new_color->b;
	bw_color = bw_color / 3;
	new_color->r = bw_color;
	new_color->g = bw_color;
	new_color->b = bw_color;
}

void portal_filter(intersect_data *id, color *new_color)
{
	intersect_data new_id;
	vector bend;
	point portal;

	CHECK_STEPS(id->step, main_scene->max_recurs);
	fill_in_next_hit_data(&new_id, id);
	
	portal.x = id->obj->amb.r;
	portal.y = id->obj->amb.g;
	portal.z = id->obj->amb.b;
	
	//vector from portal to track point (will be new proj)
	make_vector(&new_id.proj, &main_scene->camera->lookP, &portal);
	normalize(&new_id.proj);

	//next start point
	copy_vector(&new_id.start, &portal);

	//amount the current proj differs from track point vector
	make_vector(&bend, &id->proj, &main_scene->camera->look);
	normalize(&bend);

	//add amount of differ to new projection
	add_vectors(&bend, &new_id.proj, &bend);

	new_id.obj_num = trace(&new_id.start, &new_id.intersect, &new_id.proj, new_id.step);
	new_id.obj = main_scene->models[new_id.obj_num];
	new_id.obj->shader(&new_id, new_color);
}

void lens_filter(intersect_data *id, color *new_color)
{
	intersect_data new_id;
	vector bend;
	
	CHECK_STEPS(id->step, main_scene->max_recurs);
	fill_in_next_hit_data(&new_id, id);
	
	copy_vector(&new_id.start, &id->intersect);
	copy_vector(&new_id.proj, &id->proj);
	
	make_vector(&bend, &id->intersect, &id->pos);
	multiply_vector(&bend, &bend, 0.01);
	add_vectors(&new_id.proj, &new_id.proj, &bend);
	 
	new_id.start.x -= id->normal.x*0.01;
	new_id.start.y -= id->normal.y*0.01;
	new_id.start.z -= id->normal.z*0.01;
	 
	new_id.obj_num = trace(&new_id.start, &new_id.intersect, &new_id.proj, new_id.step);
	new_id.obj = main_scene->models[new_id.obj_num];
	new_id.obj->shader(&new_id, new_color);
}

void sin_filter(intersect_data *id, color *new_color)
{
	intersect_data new_id;
	vector bend;
	double dist = 0.0;

	CHECK_STEPS(id->step, main_scene->max_recurs);
	fill_in_next_hit_data(&new_id, id);
	
	copy_vector(&new_id.start, &id->intersect);
	copy_vector(&new_id.proj, &id->proj);
	
	dist = distance_between(&id->intersect, &id->obj->pos);
	make_vector(&bend, &id->intersect, &id->obj->pos);

	multiply_vector(&bend, &bend, sin(dist+current_time/TIME_SCALE));
	multiply_vector(&bend, &bend, 0.2);
	add_vectors(&new_id.proj, &new_id.proj, &bend);

	new_id.start.x -= id->normal.x*0.01;
	new_id.start.y -= id->normal.y*0.01;
	new_id.start.z -= id->normal.z*0.01;

	new_id.obj_num = trace(&new_id.start, &new_id.intersect, &new_id.proj, new_id.step);
	new_id.obj = main_scene->models[new_id.obj_num];
	new_id.obj->shader(&new_id, new_color);
}


void wire(intersect_data *i_data, color* final_color)
{
	//this shader probably doesn't work anymore- doesn't handle recursion data correctly
	color new_color;
	new_color.r = 0;
	new_color.g = 0;
	new_color.b = 0;
	intersect_data new_id;
	
	CHECK_STEPS(i_data->step, main_scene->max_recurs);
	fill_in_next_hit_data(&new_id, i_data);
	
	get_normal(&i_data->intersect, i_data->obj, &i_data->normal);
	normalize(&i_data->reflect);

	//shoot rays through the middle of the disc
	if(distance_between(&i_data->intersect, &i_data->obj->pos) < i_data->obj->radius - 0.06)
	{
		// move the intersection for internal object calculations 
		i_data->intersect.x -= i_data->normal.x*0.05;
		i_data->intersect.y -= i_data->normal.y*0.05;
		i_data->intersect.z -= i_data->normal.z*0.05;
		
		// transparency
		clear_color(&new_color);
		clear_color(final_color);
		trans_color(i_data, &new_color);
		add_colors(final_color, &new_color);
		clear_color(&new_color);
		i_data->step--;
	}
	else  //only the edges show
	{
		// phong 
		phong(i_data, &new_color);
		add_colors(final_color, &new_color);

		// reflection 	
		reflect_color(i_data, &new_color);
		multiply_color(final_color, (1 - i_data->obj->reflect));
		multiply_color(&new_color, i_data->obj->reflect);
		add_colors(final_color, &new_color);
		clear_color(&new_color);
		i_data->step--;
	}
}

void shader_template(intersect_data *hit_data_for_this_shader, color* output_from_this_shader)
{
	intersect_data hit_data_for_next_shader;
	
	CHECK_STEPS(hit_data_for_this_shader->step, main_scene->max_recurs);
	fill_in_next_hit_data(&hit_data_for_next_shader, hit_data_for_this_shader);
}

