#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "../common/objects.h"
#include "../common/vector.h"
#include "../intersect/intersect.h"
#include "../common/structs.h"
#include "../common/debug.h"
#include "./shaders.h"
#include "./lighting.h"

#include <pthread.h>
extern pthread_mutex_t mutexor;
extern scene_data* main_scene;

double ambient_occlusion(intersect_data *hit_data)
{
#define OCCLUSION_SAMPLES 200
#define MAX_OCCL_DIS 400.0
	int rays_tried;
	int check_every = 15;
	int next_check = check_every;
	double epsilon = 0.001;
	double occlusion_amount = 0;
	double occlusion_percent = 1.0;
	double last_occlusion_percent = occlusion_percent;
	vector occl_dir;
	vector occl_hit;
	vector *hit_point = &hit_data->intersect;

	for(rays_tried = 1; rays_tried <= OCCLUSION_SAMPLES; rays_tried++)
	{
		//perturb the normal
		copy_vector(&occl_dir, &hit_data->normal);
		occl_dir.x -= ((double)rand()/RAND_MAX) * 2 - 1;
		occl_dir.y -= ((double)rand()/RAND_MAX) * 2 - 1;
		occl_dir.z -= ((double)rand()/RAND_MAX) * 2 - 1;
		normalize(&occl_dir);

		//check distance
		intersect_all_objects_getptr(hit_point, &occl_dir, &occl_hit, 1);
		double occl_dis = distance_between(hit_point, &occl_hit);
		if(occl_dis > MAX_OCCL_DIS)
			occlusion_amount += 1.0;
		else
			occlusion_amount += occl_dis/MAX_OCCL_DIS;
			
		if(rays_tried == next_check)
		{
			occlusion_percent = occlusion_amount / (double)rays_tried;
			if(last_occlusion_percent == occlusion_percent)
				return occlusion_percent;

			if(fabs(last_occlusion_percent - occlusion_percent) < epsilon)
				return occlusion_percent;
			last_occlusion_percent = occlusion_percent;
			next_check += check_every;
		}
	}

	return occlusion_amount / (double)rays_tried;
}

void ambient_light(color* c, intersect_data* id, int ln)
{
	double darkener = 1;
	//vector uv;

	if(id->obj->obj_type == SPHERE)
	{
//		id->obj->inverse_map(&id->intersect, id->obj, &uv);
//		if((uv.x > 0.25 && uv.y > 0.25) || (uv.x < 0.25 && uv.y < 0.25))
//		if(uv.y > 0.5 )
//			darkener = 0.1;
	}
	
	darkener = ambient_occlusion(id);
	
	c->r += (id->obj->amb.r * main_scene->lights[ln]->amb.r * darkener);
	c->g += (id->obj->amb.g * main_scene->lights[ln]->amb.g * darkener);
	c->b += (id->obj->amb.b * main_scene->lights[ln]->amb.b * darkener);
}


void diffuse_fast(color* c, intersect_data* id, int ln)
{
	//point *intersectP = &id->intersect;
	double dpLN;
	double atten;
	double distance;
	vector lightV;
	
	double darkener = 1;
	//vector uv;
	
	sub_vectors(&lightV, &main_scene->lights[ln]->pos, &id->intersect);
	normalize(&lightV);
	dpLN = dot_product(&lightV,normalize(&id->normal));
	
	// directly perpendicular/behind object
	if (dpLN <= 0)
		dpLN = 0;
	
	/* light distance */
	distance = distance_between(&main_scene->lights[ln]->pos,
								&id->intersect);
	
	atten = 100/(50+1*distance+0.0004*pow(distance,2));
	//atten = 60/pow(distance,2);
	//atten = 3/distance;
	//atten = 0.08;
	if(main_scene->lights[ln]->obj_type == LIGHT_INFINITE)
		atten = 1;

	/*
	if(id->obj->obj_type == SPHERE)
	{
		id->obj->inverse_map(&id->intersect, id->obj, &uv);
				if((uv.x > 0.25 && uv.y > 0.25) || (uv.x < 0.25 && uv.y < 0.25))
					darkener = 3.0;
	}*/
	
	c->r += (id->obj->diff.r * dpLN *
			 main_scene->lights[ln]->diff.r * atten * darkener);
	c->g += (id->obj->diff.g * dpLN *
			 main_scene->lights[ln]->diff.g * atten * darkener);
	c->b += (id->obj->diff.b * dpLN *
			 main_scene->lights[ln]->diff.b * atten * darkener);
}

void diffuse_light(color* c, intersect_data* id, int ln)
{
	vector lightV;  // vector to light from point q
	point *intersectP = &id->intersect;
	//double dpLN;
	//double atten;
	//double distance;

	//double darkener = 1;
	//vector uv;

	/* light vector */
	lightV.x = main_scene->lights[ln]->pos.x - intersectP->x;
	lightV.y = main_scene->lights[ln]->pos.y - intersectP->y;
	lightV.z = main_scene->lights[ln]->pos.z - intersectP->z;
	normalize(&lightV);

	diffuse_fast(c, id, ln);

}
// }}}

void specular_light(color* c, intersect_data* id, int ln)
{
	vector lightV;  // vector to light from point q
	vector light_refl;
	double dpLN;
	double dpCR;
	double atten;
	double spec_gamma = id->obj->spec_alpha;
	double distance;

	lightV.x = main_scene->lights[ln]->pos.x - id->intersect.x;
	lightV.y = main_scene->lights[ln]->pos.y - id->intersect.y;
	lightV.z = main_scene->lights[ln]->pos.z - id->intersect.z;
	normalize(&lightV);
	dpLN = dot_product(&lightV,normalize(&id->normal));
	//invert_vector(&lightV);

	if (dpLN <= 0)
		dpLN = 0;
	// light distance
	distance = distance_between(&main_scene->lights[ln]->pos,
			&id->intersect);

	atten = 1/(50+2*distance+1.5*pow(distance,2))*80;
	//atten = 1/pow(distance,2)*80;
	///atten = 1/distance*20;
	//atten = 5/distance;
	//atten = 1;

	//light reflection vector
	light_refl = *reflect_vector(&light_refl, &lightV, &id->normal);

	normalize(&id->proj);
	dpCR = dot_product(&id->proj, &light_refl);
	if(dpCR < 0)
		dpCR = 0;

	c->r += (id->obj->spec.r * main_scene->lights[ln]->spec.r *
	 pow(dpCR,spec_gamma) * atten) * dpLN;
	c->g += (id->obj->spec.g * main_scene->lights[ln]->spec.g *
	 pow(dpCR,spec_gamma) * atten) * dpLN;
	c->b += (id->obj->spec.b * main_scene->lights[ln]->spec.b *
			pow(dpCR,spec_gamma) * atten) * dpLN;

}


void phong(intersect_data *id, color *new_color)
{
//	double visible = 1;
	int i = 0;
	color a,d,s,shadowed_color;
	clear_color(&a); clear_color(&d); clear_color(&s);
	clear_color(&shadowed_color);
	
	for(i=0; main_scene->lights[i]!=NULL; i++)
	{
		ambient_light(&a, id, i);
		id->other_obj_num = i;
		get_shadow(id, &shadowed_color);

		if ( shadowed_color.r > 0)
		{
			diffuse_fast(&d, id, i);
			specular_light(&s, id, i);
		}
				
		multiply_color(&d, shadowed_color.r*0.8);
		multiply_color(&s, shadowed_color.r);
		add_colors(new_color, &a);
		add_colors(new_color, &d);
		add_colors(new_color, &s);
	}
	multiply_color(new_color, 1.0/765.0);
	
	return;
}
